package org.elasticsearch.plugin.river.hazelcast

import org.elasticsearch.river._
import org.elasticsearch.client.Client
import com.hazelcast.client.{HazelcastClient, ClientConfig}
import akka.actor.{PoisonPill, ActorRef, ActorSystem}
import org.elasticsearch.common.xcontent.support.XContentMapValues._
import scala.collection.mutable.ArrayBuffer
import org.elasticsearch.common.inject.Inject
import org.elasticsearch.indices.IndexAlreadyExistsException

/** @author Stephen Samuel */
class HazelcastDriver @Inject()(riverName: RiverName,
                                settings: RiverSettings,
                                @RiverIndexName riverIndexName: String,
                                elasticClient: Client)
  extends AbstractRiverComponent(riverName, settings) with River {

    logger.debug(s"Creating river $riverName")

    import HazelcastDriver._

    val _hostname = nodeStringValue(extractValue("hazelcast.hostname", settings.settings), DefaultHazelcastHostname)
    val _port = nodeStringValue(extractValue("hazelcast.port", settings.settings), DefaultHazelcastPort)
    val _queues = Option(nodeStringValue(extractValue("hazelcast.queue", settings.settings), null)).map(_.split(",").toSeq).getOrElse(Nil)
    val _topics = Option(nodeStringValue(extractValue("hazelcast.topic", settings.settings), null)).map(_.split(",").toSeq).getOrElse(Nil)
    val _index = nodeStringValue(extractValue("index.name", settings.settings), DefaultHazelcastIndex)

    logger.debug(s"Hazelcast settings [hostname=${_hostname}, port=${_port}]")
    logger.debug(s"River settings [indexName=${_index}]")

    var system: ActorSystem = _
    var indexer: ActorRef = _
    val threads = ArrayBuffer[Thread]()
    val subscribers = ArrayBuffer[HazelcastTopicSubscriber]()

    def _getClient(hostname: String, port: String) = {
        val clientConfig = new ClientConfig()
        clientConfig.addAddress(s"$hostname:$port")
        HazelcastClient.newHazelcastClient(clientConfig)
    }

    def _ensureIndex(index: String, elasticClient: Client) {
        logger.debug("Creating index [{}]", index)
        try {
            elasticClient.admin().indices().prepareCreate(index).execute()
        } catch {
            case e: IndexAlreadyExistsException => // no problemo
        }
    }

    def start() {
        logger.debug("Starting hazelcast driver")

        system = ActorSystem("elasticsearch-river-hazelcast-actorsystem")
        logger.debug(s"Created actor system=$system")

        val hazelcastClient = _getClient(_hostname, _port)
        logger.debug(s"Created hazelcast client=$hazelcastClient")

        indexer = system.actorOf(HazelcastIndexer.props(_index, elasticClient), "indexer")
        logger.debug(s"Created indexer actor=$indexer")

        _ensureIndex(_index, elasticClient)
        _startQueues(_queues, hazelcastClient)
        _startTopics(_topics, hazelcastClient)
    }

    def _startQueues(queues: Iterable[String], hazelcastClient: HazelcastClient) {
        queues.foreach(q => {
            logger.debug(s"Consuming queue [$q]")
            val consumer = new HazelcastQueueConsumer(hazelcastClient.getQueue(q), indexer)
            val thread = new Thread(consumer)
            threads.append(thread)
            thread.start()
        })
    }

    def _startTopics(topics: Iterable[String], hazelcastClient: HazelcastClient) {
        topics.foreach(t => {
            logger.debug(s"Subscribing to topic [$t]")
            val subscriber = new HazelcastTopicSubscriber(hazelcastClient.getTopic(t), indexer)
            subscribers.append(subscriber)
        })
    }

    def close() {
        logger.debug("shutting down hazelcast river...")
        _shutdownIndexer()
        _shutdownThreads()
        _shutdownSubscribers()
        if (system != null)
            system.shutdown()
        logger.debug("...shutdown")
    }

    def _shutdownSubscribers() {
        subscribers.foreach(_.shutdown())
        subscribers.clear()
    }

    def _shutdownThreads() {
        threads.foreach(_.interrupt())
        threads.clear()
    }

    def _shutdownIndexer(): Unit = if (indexer != null) indexer ! PoisonPill
}

object HazelcastDriver {
    val DefaultHazelcastHostname = "localhost"
    val DefaultHazelcastPort = "5701"
    val DefaultHazelcastIndex = "hazelcast-index"
}